from __future__ import unicode_literals

import os

from ebb_lint.checkers.registration import register_checker
from ebb_lint.errors import Errors


@register_checker("""

( classdef< which='class' any* >
| funcdef< which='def' any* >
)

""", pass_filename=True)
def check_disallowed_dunder_init_statements(filename, which):
    if os.path.basename(filename) != '__init__.py':
        return
    yield which, Errors.no_definition_statements_in_dunder_init, {}



from __future__ import unicode_literals

from lib2to3.pgen2 import token

from ebb_lint.checkers.registration import register_checker
from ebb_lint.errors import Errors


@register_checker("""

( import_from< 'from' pdb='pdb' any* >
| import_name< 'import' pdb='pdb' any* >
| dotted_as_name< pdb='pdb' any* >
)

""")
def check_for_pdb(pdb):
    yield pdb, Errors.no_debuggers, {}


@register_checker("""

power< any+ trailer< '.' func='set_trace' > trailer< '(' ')' > >

""")
def check_for_set_trace(func):
    yield func, Errors.no_debuggers, {}


@register_checker("""

f=file_input< any* >

""", comments_for=['f'])
def scan_top_level_comments(f, f_comments):
    if '# I sincerely swear that this is one-off code.' in f_comments:
        f.print_lint_ok = True
    return []


def scan_ancestry_for(node, attr, default):
    not_set = object()
    while node is not None:
        value = getattr(node, attr, not_set)
        if value is not not_set:
            return value
        node = node.parent
    return default


@register_checker("""

( simple_stmt< any* p='print' any* >
| print_stmt< p='print' any* >
| power< p='print' trailer< '(' any* ')' > any* >
)

""")
def check_for_print(p):
    if scan_ancestry_for(p, 'print_lint_ok', False):
        return
    yield p, Errors.no_print, {}


class NoParentheizedGroup(Exception):
    pass


def parenthesized_group_leaves(container):
    if container is None or not container.children:
        raise NoParentheizedGroup()
    first_child = container.children[0]
    last_child = container.children[-1]
    if not ((not first_child.children and first_child.value == '(')
            and (not last_child.children and last_child.value == ')')):
        raise NoParentheizedGroup()
    for child in container.children[1:-1]:
        for subchild in child.pre_order():
            if not subchild.children:
                yield subchild


@register_checker("""

atom=atom< first_string=STRING STRING+ >

""")
def check_for_unintentional_implicit_concatenation(atom, first_string):
    try:
        for child in parenthesized_group_leaves(atom.parent):
            if child.type != token.STRING:
                break
        else:
            # all leaves were string literals
            return
    except NoParentheizedGroup:
        pass
    yield first_string, Errors.no_unintentional_implicit_concatenation, {}



from __future__ import unicode_literals

import collections

import enum


Error = collections.namedtuple('Error', ['code', 'message'])


class Errors(enum.Enum):
    docstrings_category = Error(
        100, 'docstring-related errors')
    no_docstring_on_init = Error(
        101,
        ("__init__ doesn't need a docstring; standard sphinx convention is to "
         "use the class-level docstring instead."))
    docstring_formatting_error = Error(
        102,
        'tried to validate the docstring: {parse_error}')
    test_docstring_prefix = Error(
        103,
        "don't start test docstrings with {prefix!r}")
    use_napoleon_in_docstrings = Error(
        104,
        'use sphinx.ext.napoleon instead of reST fields like :{field}:')

    dubious_syntax_category = Error(
        200, 'errors related to dubious syntax')
    no_trailing_comma_in_literal = Error(
        201,
        ('container literals must have a trailing comma following the last '
         'element'))
    no_print = Error(
        202,
        'print is only allowed while debugging; use logging in real code')
    no_debuggers = Error(
        203,
        ('debuggers are only allowed while debugging; make sure to remove '
         'this before committing'))
    no_unintentional_implicit_concatenation = Error(
        204,
        ('implicit string literal concatenation requires all string literals '
         'to be enclosed in parentheses containing nothing but string '
         'literals'))
    no_definition_statements_in_dunder_init = Error(
        205,
        'do not define functions or classes in __init__.py')



from __future__ import unicode_literals

from lib2to3.pygram import python_symbols

from six.moves import reduce

from ebb_lint.checkers.registration import register_checker
from ebb_lint.errors import Errors


def last(seq):
    return reduce(lambda l, r: r, seq)


@register_checker("""

( atom< '(' contents=any+ end=')' >
| atom< '[' contents=any+ end=']' >
| atom< '{' contents=any+ end='}' >
)

""")
def check_trailing_commas(end, contents):
    last_element = contents[-1]
    if (last_element.children
            and last_element.children[-1].type == python_symbols.comp_for):
        # It's a comprehension, so ignore it.
        return
    last_element_leaf = last(last_element.pre_order())
    if last_element_leaf.value == ',':
        return
    if end.value == ')' and (
            len(contents) != 1
            or contents[0].type != python_symbols.testlist_gexp):
        # It's not a tuple or a generator expression; it's just something in
        # parentheses, so ignore it.
        return
    if last_element_leaf.lineno != end.lineno:
        yield last_element_leaf, Errors.no_trailing_comma_in_literal, {}



from __future__ import unicode_literals

import collections
import re

import enum

from ebb_lint.checkers.registration import register_checker
from ebb_lint.errors import Errors


_single_line_docstring = re.compile(r'\A"""(\s*)(.*?)(\s*)"""\Z')

_docstring_leading_indent = re.compile(r'^( *)(.*)$')

_bad_test_docstring_starts = re.compile(r"""
    (?xi)
    \A(?:test|verify?|ensure?)(?:s|ies|ing)?\b
""")

_sphinx_info_fields = re.compile(r"""
    (?x)
    ^:
     (arg(?:ument)?
     |[ci]var
     |except(?:ion)?
     |key(?:word)?
     |param(?:eter)?
     |r(?:aises?|eturns?|type)
     |type
     |var(?:type)?
     )[ ]
""")


class DocstringPurpose(enum.Enum):
    test = 'test'
    other = 'other'


class Line(collections.namedtuple('Line', ['pos', 'line'])):
    def error(self, error, **extra):
        extra.update(self.pos)
        return error, extra


def docstring_errors(purpose, docstring):
    raw_lines = docstring.splitlines()
    if len(raw_lines) == 1:
        m = _single_line_docstring.match(raw_lines[0])
        if m is None:
            yield Errors.docstring_formatting_error, {
                'parse_error': (
                    'single-line docstrings must still start and end with '
                    '"""'),
            }
            return

        if not m.group(2):
            yield Errors.docstring_formatting_error, {
                'parse_error': 'single-line docstrings must not be empty',
            }
            return
        else:
            if m.group(1):
                yield Errors.docstring_formatting_error, {
                    'parse_error': (
                        'single-line docstrings must not start with '
                        'whitespace'),
                    'offset': m.start(1),
                }
            if m.group(3):
                yield Errors.docstring_formatting_error, {
                    'parse_error': (
                        'single-line docstrings must not end with whitespace'),
                    'offset': m.start(3),
                }

        lines = [
            Line({'offset': m.start(2)}, m.group(2)),
        ]

    elif len(raw_lines) < 3:
        yield Errors.docstring_formatting_error, {
            'parse_error': (
                'docstrings must have """, their text, and """ all on '
                'separate lines, or all on the same line'),
        }
        return

    else:
        if raw_lines[0] != '"""':
            yield Errors.docstring_formatting_error, {
                'parse_error': (
                    "a multi-line docstring's first line must be just " '"""'),
            }
            return
        initial_indent = _docstring_leading_indent.match(raw_lines[1])
        indent = initial_indent.end(1)
        if indent % 4 != 0:
            yield Errors.docstring_formatting_error, {
                'parse_error': (
                    'docstring indentation must be a multiple of four'),
                'line_offset': 1,
                'column': 0,
            }
        first_line = initial_indent.group(2)
        if not first_line:
            yield Errors.docstring_formatting_error, {
                'parse_error': (
                    'multi-line docstrings must have text on the first line '
                    'following the """'),
                'line_offset': 1,
                'column': indent,
            }
            return

        indent_string = initial_indent.group(1)
        lines = [first_line]
        all_lines_good = True
        for line_offset, raw_line in enumerate(raw_lines[2:], start=2):
            if not raw_line:
                lines.append(None)
            elif raw_line.startswith(indent_string):
                lines.append(raw_line[indent:])
            else:
                yield Errors.docstring_formatting_error, {
                    'parse_error': (
                        'every line of a docstring must have the same '
                        'indentation'),
                    'line_offset': line_offset,
                    'column': 0,
                }
                all_lines_good = False
                lines.append(None)

        if all_lines_good and lines[-1] != '"""':
            yield Errors.docstring_formatting_error, {
                'parse_error': 'multi-line docstrings must end with """',
                'line_offset': len(raw_lines) - 1,
                'column': indent,
            }

        lines = [
            Line({'line_offset': line_offset, 'column': indent}, line)
            for line_offset, line in enumerate(lines, start=1)]

    if purpose == DocstringPurpose.test:
        m = _bad_test_docstring_starts.match(lines[0].line)
        if m is not None:
            yield lines[0].error(
                Errors.test_docstring_prefix, prefix=m.group())

    for line in lines:
        if line.line is None:
            continue
        m = _sphinx_info_fields.match(line.line)
        if m is not None:
            yield line.error(
                Errors.use_napoleon_in_docstrings, field=m.group(1))


@register_checker(r"""

( classdef< which='class' name=NAME any*
            suite< '\n' TOKEN simple_stmt< docstring=STRING any > any* > >
| funcdef< which='def' name=NAME any*
           suite< '\n' TOKEN simple_stmt< docstring=STRING any > any* > >
)

""")
def check_docstring(which, name, docstring):
    purpose = DocstringPurpose.other
    if which.value == 'def':
        if name.value.startswith('test_'):
            purpose = DocstringPurpose.test
        elif name.value == '__init__':
            yield docstring, Errors.no_docstring_on_init, {}
    for error, kw in docstring_errors(purpose, docstring.value):
        yield docstring, error, kw



from __future__ import unicode_literals

import bisect
import io
import tokenize
from lib2to3.pgen2 import driver, token
from lib2to3 import patcomp, pygram, pytree

import six
import venusian

from ebb_lint._version import __version__
from ebb_lint import checkers


# Stolen from lib2to3 directly. Why was this a private function? Ugh.
def detect_future_features(infile):
    have_docstring = False
    gen = tokenize.generate_tokens(infile.readline)

    def advance():
        tok = next(gen)
        return tok[0], tok[1]

    ignore = frozenset((token.NEWLINE, tokenize.NL, token.COMMENT))
    features = set()
    try:
        while True:
            tp, value = advance()
            if tp in ignore:
                continue
            elif tp == token.STRING:
                if have_docstring:
                    break
                have_docstring = True
            elif tp == token.NAME and value == 'from':
                tp, value = advance()
                if tp != token.NAME or value != '__future__':
                    break
                tp, value = advance()
                if tp != token.NAME or value != 'import':
                    break
                tp, value = advance()
                if tp == token.OP and value == '(':
                    tp, value = advance()
                while tp == token.NAME:
                    features.add(value)
                    tp, value = advance()
                    if tp != token.OP or value != ',':
                        break
                    tp, value = advance()
            else:
                break
    except StopIteration:
        pass
    return frozenset(features)


if six.PY3:
    def grammar_for_filename(filename):
        return pygram.python_grammar_no_print_statement

else:
    def grammar_for_filename(filename):
        with open(filename, 'r') as infile:
            future_features = detect_future_features(infile)
        if 'print_function' in future_features:
            return pygram.python_grammar_no_print_statement
        else:
            return pygram.python_grammar


def find_comments(s):
    fobj = io.StringIO(six.text_type(s))
    for typ, tok, _, _, _ in tokenize.generate_tokens(fobj.readline):
        if typ == tokenize.COMMENT:
            yield tok


class Lines(object):
    def __init__(self, infile):
        count = 0
        self.lines = [(0, '')]
        for line in infile:
            self.lines.append((count, line))
            count += len(line)

    def __getitem__(self, idx):
        return self.lines[idx]

    def position_of_byte(self, byte):
        lineno = bisect.bisect_left(self.lines, (byte + 1,)) - 1
        column = byte - self.lines[lineno][0]
        return lineno, column


class EbbLint(object):
    name = 'ebb_lint'
    version = __version__

    _lines = None

    def __init__(self, tree, filename):
        self.tree = tree
        self.filename = filename

    @classmethod
    def add_options(cls, parser):
        pass

    @classmethod
    def parse_options(cls, options):
        collected_checkers = []

        def register_checker(pattern, checker, extra):
            pattern = patcomp.compile_pattern(pattern)
            collected_checkers.append((pattern, checker, extra))

        scanner = venusian.Scanner(register=register_checker)
        scanner.scan(checkers)
        cls.collected_checkers = collected_checkers

    @property
    def lines(self):
        if self._lines is None:
            with open(self.filename) as infile:
                self._lines = Lines(infile)
        return self._lines

    def _message_for(self, node, error, **kw):
        line_offset = kw.pop('line_offset', None)
        if line_offset is None:
            byte, _ = self.lines[node.lineno]
            byte += node.column + kw.pop('offset', 0)
            lineno, column = self.lines.position_of_byte(byte)
        else:
            lineno = node.lineno + line_offset
            column = kw.pop('column')
        message = 'L{:03d} {}'.format(
            error.value.code, error.value.message.format(**kw))
        return lineno, column, message, type(self)

    def run(self):
        d = driver.Driver(
            grammar_for_filename(self.filename), convert=pytree.convert)
        tree = d.parse_file(self.filename)
        for node in tree.pre_order():
            for pattern, checker, extra in self.collected_checkers:
                results = {}
                if not pattern.match(node, results):
                    continue
                for k in extra.get('comments_for', ()):
                    comments = list(find_comments(node.prefix))
                    results[k + '_comments'] = comments
                if extra.get('pass_filename', False):
                    results['filename'] = self.filename
                for error_node, error, kw in checker(**results):
                    yield self._message_for(error_node, error, **kw)



import venusian


def register_checker(pattern, **extra):
    def deco(func):
        def callback(scanner, name, obj):
            scanner.register(pattern, obj, extra)
        venusian.attach(func, callback)
        return func
    return deco



from __future__ import unicode_literals

import ast
import re

import pytest
import six

from ebb_lint.flake8 import EbbLint, Lines


py3skip = pytest.mark.skipif(six.PY3, reason='not runnable on python 3')


_code_pattern = re.compile(r"""
    (?x)
    \$([A-Z]\d{3})\$
""")


def find_error_locations(source):
    error_locations = []

    def replacement(match):
        last_offset = error_locations[-1][2] if error_locations else 0
        new_offset = last_offset + match.end() - match.start()
        error_locations.append((
            match.start() - last_offset, match.group(1), new_offset))
        return ''

    source = _code_pattern.sub(replacement, source)
    lines = Lines(source.splitlines(True))
    return source, [
        lines.position_of_byte(b) + (code,) for b, code, _ in error_locations]


@pytest.fixture(autouse=True, scope='session')
def scan_for_checkers():
    EbbLint.add_options(None)
    EbbLint.parse_options(None)


all_sources = [
    """

def f():
    $L102$'''
    spam
    '''

    """,

    """

def f():
    $L102$'''spam'''

    """,

    """

def f():
    $L102$'spam'

    """,

    """

def f():
    $L102$"spam"

    """,

    '''

def f():
    """$L102$ spam"""

    ''',

    '''

def f():
    """spam$L102$ """

    ''',

    '''

def __init__():
    $L101$"""
    spam
    """

    ''',

    '''

class Spam(object):
    """
    It's spam.
    """

    ''',

    '''

import $L203$pdb
pdb.$L203$set_trace()

    ''',

    '''

import $L203$pdb as not_pdb
not_pdb.$L203$set_trace()

    ''',

    '''

from $L203$pdb import set_trace
set_trace()

    ''',

    '''

import pudb
pudb.$L203$set_trace()

    ''',

    '''

x = $L204$'a' 'b'

    ''',

    '''

x = [
    $L204$'a'
    'b',
    'c',
]

    ''',

    '''

x = [
    ($L204$'a'
     'b'.format()),
    'c',
]

    ''',

    '''

x = [
    ($L204$'a'
     'b'()),
    'c',
]

    ''',

    '''

x = [
    [$L204$'a'
     'b'],
    'c',
]

    ''',

    '''

x = [
    ('a'
     'b').format(),
    'c',
]

    ''',

    '''

x = [
    ('a'
     'b'),
    'c',
]

    ''',

    '''

x = ('a' 'b')

    ''',

    '''

from __future__ import print_function
import sys

$L202$print("Yay!", file=sys.stdout)

    ''',

    '''

# I sincerely swear that this is one-off code.
from __future__ import print_function
import sys

print("Yay!", file=sys.stdout)

    ''',
]


contexts = [
    ('# I sincerely swear that this is one-off code.', ''),

    ("""
# I sincerely swear that this is one-off code.
# I sincerely swear that I am not a member of the Communist Party.
    """, ''),

    ('', '$L202$'),
]


all_sources.extend(
    py3skip('''
{ctx[0]}
{ctx[1]}{print_}
    '''.format(ctx=ctx, print_=print_))
    for ctx in contexts
    for print_ in [
        "print 'hi'",
        "print 'hi',",
        "print >> aether, 'hi'",
        "print",
    ])


all_sources.extend(
    '''
{ctx[0]}
from __future__ import print_function
{ctx[1]}{print_}
    '''.format(ctx=ctx, print_=print_)
    for ctx in contexts
    for print_ in [
        "print('hi')",
        "print()",
    ])


element_pairs = [
    ('1', '2'),
    ("'spam'", "'eggs'"),
]


delimiter_pairs = [
    (('(', ')'), ['f(x)']),
    (('[', ']'), ['f(x)']),
    (('{', '}'), ['f(x)', 'f(x): g(x)']),
]


all_sources.extend(
    template.format(delim=delim, elem=elem)
    for delim, _ in delimiter_pairs
    for elem in element_pairs
    for template in [
        '''

spam = {delim[0]}
    {elem[0]},
    {elem[1]},
{delim[1]}

        ''',

        '''

spam = {delim[0]}
    {elem[0]},
{delim[1]}

        ''',

        '''

spam = {delim[0]}
    {elem[0]},
    $L201${elem[1]}
{delim[1]}

        ''',

        '''

spam = {delim[0]}{elem[0]}, {elem[1]}{delim[1]}

        ''',

        '''

spam = {delim[0]}{elem[0]}{delim[1]}

        ''',
    ])


all_sources.extend(
    '''

spam = {delim[0]}
    $L201${elem[0]}
{delim[1]}

    '''.format(delim=delim, elem=elem)
    for delim, _ in delimiter_pairs
    for elem in element_pairs
    # Ignore this because it's a parenthesized expression, and not a tuple.
    if delim[0] != '(')


all_sources.extend(
    template.format(delim=delim)
    for delim, _ in delimiter_pairs
    for template in [
        '''

spam = {delim[0]}{delim[1]}

        ''',

        '''

spam = {delim[0]}
{delim[1]}

        ''',
    ])


all_sources.extend(
    template.format(delim=delim, expr=expr)
    for delim, exprs in delimiter_pairs
    for expr in exprs
    for template in [
        '''

spam = {delim[0]}
    {expr} for x in y
{delim[1]}

        ''',

        '''

spam = {delim[0]}
    {expr}
    for x in y
{delim[1]}

        ''',

        '''

spam = {delim[0]}
    {expr}
    for x in y
    if p(x)
{delim[1]}

        ''',

        '''

spam = {delim[0]}
    {expr}
    for y in z
    for x in y
    if p(x)
{delim[1]}

        ''',

        '''

spam = {delim[0]}
    {expr} for x in y{delim[1]}

        ''',

        '''

spam = {delim[0]}{expr} for x in y{delim[1]}

        ''',
    ])


dict_element_pairs = [
    (('1', '2'), ('3', '4')),
    (("'spam'", "'eggs'"), ("'eggs'", "'spam'")),
]


all_sources.extend(
    template.format(elem=elem)
    for elem in dict_element_pairs
    for template in [
        '''

spam = {{
    {elem[0][0]}: {elem[0][1]},
    {elem[1][0]}: {elem[1][1]},
}}

        ''',

        '''

spam = {{
    {elem[0][0]}: {elem[0][1]},
    {elem[1][0]}: $L201${elem[1][1]}
}}

        ''',

        '''

spam = {{{elem[0][0]}: {elem[0][1]}, {elem[1][0]}: {elem[1][1]}}}

        ''',
    ])


docstrings = [
    """
    hi
    """,

    """
    hi

    hello
    """,

    """
    hi

      hello

    hey
    """,

    """
        hi
        """,

    """
        hi

        hello
        """,

    """
        hi

          hello

        hey
        """,

    '\nhi\n',

    '\nhi\n\nhello\n',

    '\nhi\n\n  hello\n\nhey\n',

    """
    hi

$L102$  hey

    hello
    """,

    """
$L102$
    hi
    """,

    """
    $L102$
    hi
    """,

    '\n    hi\n$L102$',

    """
$L102$   hey
   """,

    """
$L102$   hey
   $L102$ """,

    'hi',
]


docstrings.extend("""
    $L104$:{} spam: eggs
    """.format(field) for field in [
    'param', 'parameter', 'arg', 'argument', 'key', 'keyword', 'type',
    'raises', 'raise', 'except', 'exception', 'var', 'ivar', 'cvar',
    'vartype', 'returns', 'return', 'rtype',
])


all_sources.extend('''

def f():
    """{}"""

'''.format(docstring) for docstring in docstrings)


all_sources.extend('''

def f():
    $L102$"""{}"""

'''.format(docstring) for docstring in [
    """
""",

    """
 """,

    """
hi""",

    """hi
    """,

    '',
])


test_docstrings = [
    """
    $L103$Test that the thing does the thing.
    """,

    '$L103$Test that the thing does the thing.',

    """
    $L103$VERIFY MY YELLING VOLUME
    """,

    """
    MY YELLING VOLUME IS FINE
    """,
]


test_docstrings.extend("""
    $L103${} a thing.
    """.format(word) for word in [
    'test', 'verify', 'ensure', 'tests', 'verifies', 'ensures', 'testing',
    'verifying', 'ensuring',
])


all_sources.extend('''

def test_f():
    """{}"""

'''.format(docstring) for docstring in test_docstrings)


dunder_init_sources = [
    '''

$L205$def spam():
    pass

    ''',

    '''

$L205$class Spam():
    pass

    ''',

    '''

from spam import eggs

__all__ = ['eggs']

    ''',
]


all_filename_sources = [
    ('__init__.py', source) for source in dunder_init_sources]


@pytest.mark.parametrize(('filename', 'source'), all_filename_sources)
def test_linting_with_filename(tmpdir, source, filename):
    source, error_locations = find_error_locations(source)
    sourcefile = tmpdir.join(filename)
    sourcefile.write(source)
    lint = EbbLint(ast.parse(source), sourcefile.strpath)
    actual = [
        (line, col, message[:4]) for line, col, message, _ in lint.run()]
    assert actual == error_locations


@pytest.mark.parametrize('source', all_sources)
def test_linting_with_default_filename(tmpdir, source):
    return test_linting_with_filename(tmpdir, source, filename='source.py')



import os
import sys

from setuptools import setup


# required due to vboxfs not supporting hard links. this doesn't have any ill
# effects. it just means there's a bit of copying instead of hard linking.
del os.link


extras_require = {
    'dev': [
        'coverage',
        'pytest',
        'pytest-cov',
    ],
}

extras_require['all'] = list({
    dep for deps in extras_require.values() for dep in deps})


install_requires = [
    'flake8',
    'six',
    'venusian',
]

if sys.version_info < (3, 4):
    install_requires.append('enum34')


with open('README.rst') as infile:
    long_description = infile.read()


setup(
    name='ebb-lint',
    vcversioner={
        'version_module_paths': ['ebb_lint/_version.py'],
    },
    description='lint for ensuring quality software',
    long_description=long_description,
    author='Flowroute Inc.',
    author_email='development@flowroute.com',
    url='https://github.com/flowroute/ebb-lint',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Plugins',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Software Development :: Quality Assurance',
    ],
    license='MIT',
    packages=[
        'ebb_lint',
        'ebb_lint.checkers',
        'ebb_lint.test',
    ],
    install_requires=install_requires,
    extras_require=extras_require,
    setup_requires=['vcversioner'],
    entry_points={
        'flake8.extension': [
            'L = ebb_lint:EbbLint',
        ],
    },
)
